# Challenges of index exchange for search engine interoperability

This repository contains the code used to run the experiments in the short paper "Challenges of index exchange for search engine interoperability", published at OSSYM 2023.

For our experiments, we index Robust04 into three different IR systems:

- [PyTerrier](https://github.com/terrier-org/pyterrier)
- [Anserini](https://github.com/castorini/anserini)
- [GeeseDB](https://github.com/informagi/GeeseDB)

Then, we exchange indexes between the different systems by exporting the index from each system into [the CIFF format](https://github.com/osirrc/ciff), and then importing it into the other two. We then show that retrieval performance on Robust04 significantly drops, due to tokenization differences between the different systems. We also propose a generic tokenizer that tokenizes content based on the dictionary contained in a CIFF file, which alleviates the observed tokenization problems.

## Installation

We use [Poetry](https://python-poetry.org/) for dependency management. To install all dependencies, simply issue the following command:

    $ poetry install

Then, run all your subsequent in the Poetry environment by running `poetry shell`:

    $ poetry shell
    (.venv) $ python3 build_indexes.py 

Or, prepend all your commands with `poetry run`:

    $ poetry run python3 build_indexes.py

This project also uses some external tools to, among other things, import and export CIFF files.

### Anserini

While we are able to use Pyserini bindings to index the corpus into Anserini, we were unable to get Pyserini working for retrieval from Anserini (while also running PyTerrier). Therefore, we use Anserini's `SearchCollection` tool to perform retrieval tasks on the Anserini index.

To install this, you can simply clone and build Anserini, and add the Anserini binaries to your PATH:

    $ git clone https://github.com/castorini/anserini.git
    $ cd anserini
    $ mvn -Dmaven.test.skip=true clean package appassembler:assemble
    $ export PATH="$PATH:$PWD/target/appassembler/bin"

### CIFF tools

To export an Anserini(/Lucene) index into a CIFF file, we use the `ExportAnseriniLuceneIndex` application in the CIFF repository:

    $ git clone https://github.com/osirrc/ciff.git
    $ cd ciff
    $ mvn -Dmaven.test.skip=true clean package appassembler:assemble
    $ export PATH="$PATH:$PWD/target/appassembler/bin"

### Lucene CIFF importer

To import a CIFF file into Anserin(/Lucene), we use [the Lucene CIFF importer](https://github.com/informagi/lucene-ciff).

    $ git clone https://github.com/informagi/lucene-ciff.git
    $ cd lucene-ciff
    $ mvn -Dmaven.test.skip=true clean package appassembler:assemble
    $ export PATH="$PATH:$PWD/target/appassembler/bin"

### Robust04

We use `ir-datasets` to access Robust04 in a structured manner. Before starting, ensure you have access to Robust04, and that you put it in a location where `ir-datasets` can find and use it.

For more information, refer to [the ir-datasets documentation](https://ir-datasets.com/disks45.html).

_Note: we encountered problems while reading the Robust04 files with .0z, .1z and .2 extensions using ir-datasets. See also [this issue](https://github.com/allenai/ir_datasets/issues/244) explaining the problem and [this pull request](https://github.com/allenai/ir_datasets/pull/247) offering a solution._

## Usage

Our experimental setup is split into two stages. First, we build all the relevant indexes, export them to CIFF files, import them into the other retrieval systems, and generate lexicons for each index. This is done using the `build_indexes.py` script.

Once all the indexes have been constructed, we can compare the performance between different systems. This is done with the `compare_effectiveness.py` script.

### Index generation

To generate the indexes, simply run the command:

    $ python3 build_indexes.py

This script will start populating an `indexes` folder, containing a directory for each of the used IR systems. The directory contains the following files:

| Directory / File                 | Description                                                                                                                                  |
|----------------------------------|----------------------------------------------------------------------------------------------------------------------------------------------|
| `indexes/[SOURCE]/[SOURCE]/`     | The Robust04 index in `[SOURCE]` format, where `[SOURCE]` can be PyTerrier, Anserini or GeeseDB                                              |
| `indexes/[SOURCE]/[SOURCE].ciff` | The CIFF export of the generated Robust04 index.                                                                                             |
| `indexes/[SOURCE]/dict.txt`      | The lexicon of the exported CIFF index.                                                                                                      |
| `indexes/[SOURCE]/[TARGET]/`     | The CIFF index imported into the `[TARGET]` format, where `[TARGET]` can be PyTerrier, Anserini or GeeseDB and is different from `[SOURCE]`. |

### Running the experiment

To compare two systems, run the `compare_effectiveness.py` script, with the names of the systems to compare. For instance:

    $ python3 compare_effectiveness.py pyterrier geesedb
    ...
    ================================================================
    Source index: Terrier
    Target index: GeeseDB
    ================================================================
    --- All Robust04 topics ----------------------------------------
                         MAP   nDCG
    System  Tokenizer              
    Terrier Terrier    0.221  0.480
    GeeseDB NLTK       0.208  0.457
            CIFF       0.224  0.482
    
    --- Robust04 topics that contain a hyphen or period ------------
                         MAP   nDCG
    System  Tokenizer              
    Terrier Terrier    0.234  0.541
    GeeseDB NLTK       0.081  0.292
            CIFF       0.234  0.541

To directly generate the LaTeX tables in our paper, add the `--to-latex` flag:

    $ python3 compare_effectiveness.py --to-latex pyterrier geesedb

This will print the same tables in LaTeX commands, including captions, labels and statistical significance markers.
