from pathlib import Path

from ciff_toolkit.read import CiffReader
import ir_datasets
import pyterrier as pt
from tqdm import tqdm

pt.init(packages=["com.github.terrierteam:terrier-ciff:-SNAPSHOT"], boot_packages=["io.anserini:anserini:0.21.0:fatjar"])

from index import Index, TerrierIndex, AnseriniIndex, GeeseDBIndex
from config import DATASET, INDEX_DIR


VERBOSE = True

INDEXES = {
    'pyterrier': TerrierIndex(verbose=VERBOSE),
    'anserini': AnseriniIndex(verbose=VERBOSE),
    'geesedb': GeeseDBIndex(verbose=VERBOSE),
}


def export_ciff_dictionary(ciff_file: Path, dict_file: Path, verbose: bool = False):
    with CiffReader(ciff_file) as reader, open(dict_file, 'w') as _file:
        for pl in tqdm(reader.read_postings_lists(), disable=not verbose):
            _file.write(f'{pl.term}\n')


def generate_indexes(name: str, index: Index, dataset: ir_datasets.Dataset):
    (INDEX_DIR / name).mkdir(parents=True, exist_ok=True)

    index_name = INDEX_DIR / name / name
    ciff_name = INDEX_DIR / name / f'{name}.ciff'
    dict_name = INDEX_DIR / name / 'dict.txt'

    if index_name.exists():
        print(f'Index for {name} already exists!')
    else:
        print(f'Indexing dataset into {name} ({index_name})...')
        index.index_collection(index_name, dataset)

    if ciff_name.exists():
        print(f'CIFF export for {name} already exists!')
    else:
        print(f'Exporting {name} index to CIFF ({ciff_name})...')
        index.export_ciff(index_name, ciff_name)

    if dict_name.exists():
        print(f'Lexicon for {name} already exists!')
    else:
        print(f'Exporting dictionary for {name} ({dict_name})...')
        export_ciff_dictionary(ciff_name, dict_name, verbose=VERBOSE)

    for target_name, target_index in INDEXES.items():
        if name == target_name:
            continue

        target_index_name = INDEX_DIR / name / target_name

        if target_index_name.exists():
            print(f'Index for {target_name}, imported from {name}, already exists!')
        else:
            print(f'Loading {name} CIFF file into {target_name} ({target_index_name})...')
            target_index.import_ciff(ciff_name, target_index_name)


def main():
    dataset = ir_datasets.load(DATASET)

    for name, index in INDEXES.items():
        generate_indexes(name, index, dataset)


if __name__ == '__main__':
    main()
