from argparse import ArgumentParser
from typing import Optional

import numpy as np
import pandas as pd
import pyterrier as pt
pt.init(boot_packages=["io.anserini:anserini:0.9.2:fatjar"])

from index import AnseriniIndex, GeeseDBIndex
from systems import AnseriniBatchRetrieve, GeeseDBBatchRetrieve, PyTerrierBatchRetrieve
from ciff_tokenizer import CiffTokenizer

from config import *


MODELS = {
    'pyterrier': PyTerrierBatchRetrieve,
    'anserini': AnseriniBatchRetrieve,
    'geesedb': GeeseDBBatchRetrieve,
}

TOKENIZERS = {
    'pyterrier': pt.rewrite.tokenise(),
    'anserini': pt.rewrite.tokenise(AnseriniIndex.tokenize),
    'geesedb': pt.rewrite.tokenise(GeeseDBIndex.tokenize),
}

MODEL_NAMES = {
    'pyterrier': 'Terrier',
    'anserini': 'Anserini',
    'geesedb': 'GeeseDB',
}

TOKENIZER_NAMES = {
    'pyterrier': 'Terrier',
    'anserini': 'Anserini',
    'geesedb': 'NLTK',
}


class Experiment:

    def __init__(self, source_system: str, target_system: str, index_dir: Path, dataset: pt.datasets.Dataset,
                 measures: tuple[str] = ('map', 'ndcg'), significance_threshold: float = 0.01, **model_kwargs) -> None:
        self.source_system = source_system
        self.target_system = target_system
        self.index_dir = index_dir
        self.dataset = dataset
        self.measures = measures
        self.significance_threshold = significance_threshold
        self.model_kwargs = model_kwargs

    def evaluate_systems(self, char_subset: Optional[str] = None, baseline: Optional[int] = None) -> pd.DataFrame:
        topics = self.dataset.get_topics(variant='title', tokenise_query=False)
        qrels = self.dataset.get_qrels()

        if char_subset is not None:
            qids = [
                qid
                for qid, query in topics[['qid', 'query']].itertuples(index=False)
                if any(c in query for c in char_subset)
            ]

            topics = topics[topics['qid'].isin(qids)]
            qrels = qrels[qrels['qid'].isin(qids)]

        names = [f'{name} ({tokenizer})' for name, tokenizer in zip(self.get_system_names(), self.get_tokenizer_names())]

        return pt.Experiment(self.get_systems(), topics, qrels, self.measures, names, baseline=baseline)

    def get_systems(self) -> list[pt.Transformer]:
        ciff_tokenizer = CiffTokenizer(self.index_dir / self.source_system / 'dict.txt')
        ciff_tokenize = pt.rewrite.tokenise(lambda x: ciff_tokenizer.tokenize(x), matchop=True)

        return [
            (TOKENIZERS[self.source_system] >>
             MODELS[self.source_system](self.index_dir / self.source_system / self.source_system, **self.model_kwargs)),
            (TOKENIZERS[self.target_system] >>
             MODELS[self.target_system](self.index_dir / self.source_system / self.target_system, **self.model_kwargs)),
            (ciff_tokenize >>
             MODELS[self.target_system](self.index_dir / self.source_system / self.target_system, **self.model_kwargs)),
        ]

    def get_system_names(self) -> list[str]:
        return [MODEL_NAMES[self.source_system], MODEL_NAMES[self.target_system], MODEL_NAMES[self.target_system]]

    def get_tokenizer_names(self) -> list[str]:
        return [TOKENIZER_NAMES[self.source_system], TOKENIZER_NAMES[self.target_system], 'CIFF']

    def run(self):
        print('=' * 64)
        print(f'Source index: {MODEL_NAMES[self.source_system]}')
        print(f'Target index: {MODEL_NAMES[self.target_system]}')
        print('=' * 64)

        print('---', 'All Robust04 topics', '-' * 40)
        print(self.format_result(self.evaluate_systems()).round(3))
        print()

        print('---', 'Robust04 topics that contain a hyphen or period', '-' * 12)
        print(self.format_result(self.evaluate_systems(char_subset=PERIOD_HYPHEN_SUBSET)).round(3))
        print()

    def to_latex(self) -> str:
        lines = [
            '\\newcommand{\\sig}[1]{#1$^\\dagger$}\n',
            '\\begin{table}\n',
            '\\caption{Performance of different systems using an inverted file',
            f'\tgenerated with {MODEL_NAMES[self.source_system]}. Best results are marked in',
            f'bold. Configurations that perform significantly ($p < {self.significance_threshold}$) better',
            f'than the {MODEL_NAMES[self.source_system]} index imported into {MODEL_NAMES[self.target_system]}',
            '(the middle row) are marked with $\\dagger$.}',

        ]

        all_data = self.evaluate_systems(baseline=1)
        lines.append(self.result_to_latex(all_data, 'All Robust04 topics',
                                          f'{self.source_system}-all-topics'))

        subset_data = self.evaluate_systems(char_subset=PERIOD_HYPHEN_SUBSET, baseline=1)
        lines.append(self.result_to_latex(subset_data, 'Robust04 topics that contain a hyphen or period',
                                          f'{self.source_system}-topic-subset'))

        lines.append('\\end{table}')

        return '\n'.join(lines)

    def result_to_latex(self, result: pd.DataFrame, caption: str, label: str) -> str:
        significances = result[['map p-value', 'ndcg p-value']] < self.significance_threshold

        return (
            self.format_result(result)
                .style
                .format(precision=3)
                .highlight_max(axis='rows', props='bfseries: ;')
                .apply(lambda _: np.where(significances, 'sig: --rwrap;', ''), axis=None)
                .to_latex(multicol_align='c', caption=caption, label=label, position_float='centering',
                        hrules=True, clines='skip-last;data', siunitx=True, environment='subtable')
                .replace('\\begin{subtable}', '\\begin{subtable}{\\linewidth}')
        )

    def format_result(self, result: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame({
            'System': self.get_system_names(),
            'Tokenizer': self.get_tokenizer_names(),
            'MAP': result['map'],
            'nDCG': result['ndcg'],
        }).set_index(['System', 'Tokenizer'])


def main():
    parser = ArgumentParser()

    parser.add_argument('--to-latex', action='store_true',
                        help='Print LaTeX tables instead of regular output')
    parser.add_argument('source', help='The source index', choices=['pyterrier', 'anserini', 'geesedb'])
    parser.add_argument('target', help='The target index', choices=['pyterrier', 'anserini', 'geesedb'])

    args = parser.parse_args()

    dataset = pt.get_dataset(f'irds:{DATASET}')

    experiment = Experiment(args.source, args.target, INDEX_DIR, dataset,
                            num_results=NUM_RESULTS, b=BM25_B, k1=BM25_K1)

    if args.to_latex:
        print(experiment.to_latex())
    else:
        experiment.run()


if __name__ == '__main__':
    main()
