import csv
import subprocess
import tempfile
from abc import ABC, abstractmethod
from collections import Counter
from pathlib import Path
from multiprocessing.pool import Pool
from typing import Iterable

import nltk
import pandas as pd
from ciff_toolkit.ciff_pb2 import PostingsList, Posting, Header, DocRecord
from ciff_toolkit.read import CiffReader
from ciff_toolkit.write import CiffWriter
from geesedb.connection import close_connection
from geesedb.index.fulltext_from_csv import FullTextFromCSV
from geesedb.utils.ciff.to_ciff import ToCiff
from ir_datasets import Dataset

import pyterrier as pt
from nltk import word_tokenize
from tqdm import tqdm

assert pt.started()


class Index(ABC):

    def __init__(self, verbose: bool = False):
        self.verbose = verbose
        self.disable_tqdm = not verbose

    @abstractmethod
    def index_collection(self, index_name: Path, dataset: Dataset):
        pass

    @abstractmethod
    def import_ciff(self, ciff_file: Path, index_name: Path):
        pass

    @abstractmethod
    def export_ciff(self, index_name: Path, ciff_file: Path):
        pass

    @staticmethod
    def doc_text(doc):
        return getattr(doc, 'text', getattr(doc, 'body', ''))


class TerrierIndex(Index):

    def index_collection(self, index_name: Path, dataset: Dataset):
        def _loader():
            for doc in tqdm(dataset.docs_iter(), total=dataset.docs_count(),
                            disable=self.disable_tqdm, desc='Indexing'):
                yield {'docno': doc.doc_id, 'text': self.doc_text(doc)}

        indexer = pt.index.IterDictIndexer(str(index_name), stopwords='none', stemmer='none', threads=8)
        indexer.index(_loader())

    def import_ciff(self, ciff_file: Path, index_name: Path):
        index_name.mkdir(parents=True, exist_ok=True)
        pt.run('ciff-ingest', ['-I', str(index_name / 'data.properties'), str(ciff_file)])

    def export_ciff(self, index_name: Path, ciff_file: Path):
        indexref = pt.IndexRef.of(str(index_name))
        index = pt.IndexFactory.of(indexref)

        with CiffWriter(ciff_file) as writer:
            writer.write_header(self.get_ciff_header(index))
            writer.write_postings_lists(self.get_ciff_postings_lists(index))
            writer.write_documents(self.get_ciff_doc_records(index))

    @staticmethod
    def get_ciff_header(index) -> Header:
        # See: https://github.com/terrierteam/terrier-ciff/blob/master/trindex_to_ciff.ipynb
        header = Header()
        header.version = 1
        header.num_postings_lists = index.getCollectionStatistics().getNumberOfUniqueTerms()
        header.num_docs = index.getCollectionStatistics().getNumberOfDocuments()
        header.total_docs = index.getCollectionStatistics().getNumberOfDocuments()
        header.total_postings_lists = index.getCollectionStatistics().getNumberOfUniqueTerms()
        header.total_terms_in_collection = index.getCollectionStatistics().getNumberOfTokens()
        header.average_doclength = index.getCollectionStatistics().getAverageDocumentLength()
        header.description = index.getIndexRef().toString()

        return header

    def get_ciff_postings_lists(self, index) -> Iterable[PostingsList]:
        # See: https://github.com/terrierteam/terrier-ciff/blob/master/trindex_to_ciff.ipynb
        lexicon = index.getLexicon()
        inv = index.getInvertedIndex()

        for kv in tqdm(lexicon, disable=self.disable_tqdm, desc='Posting lists'):
            pl = PostingsList()
            pl.term = kv.getKey()
            pl.df = kv.getValue().getDocumentFrequency()
            pl.cf = kv.getValue().getFrequency()

            prev_docid = 0
            for p in inv.getPostings(kv.getValue()):
                docid = p.getId()

                posting = Posting()
                posting.docid = docid - prev_docid
                posting.tf = p.getFrequency()

                pl.postings.append(posting)
                prev_docid = docid

            yield pl

    def get_ciff_doc_records(self, index) -> Iterable[DocRecord]:
        # See: https://github.com/terrierteam/terrier-ciff/blob/master/trindex_to_ciff.ipynb
        metaindex = index.getMetaIndex()
        docindex = index.getDocumentIndex()

        for docid in tqdm(range(index.getCollectionStatistics().getNumberOfDocuments()),
                          disable=self.disable_tqdm, desc='Documents'):
            doc = DocRecord()
            doc.docid = docid
            doc.collection_docid = metaindex.getItem("docno", docid)
            doc.doclength = docindex.getDocumentLength(docid)

            yield doc


AnalyzerUtils = pt.autoclass('io.anserini.analysis.AnalyzerUtils')
String = pt.autoclass('java.lang.String')


class AnseriniIndex(Index):

    def index_collection(self, index_name: Path, dataset: Dataset):
        SimpleIndexer = pt.autoclass('io.anserini.index.SimpleIndexer')
        JsonCollectionDocument = pt.autoclass('io.anserini.collection.JsonCollection$Document')

        indexer = SimpleIndexer(['-index', str(index_name), '-optimize', '-keepStopwords', '-stemmer', 'none',
                                 '-input', '', '-collection', 'JsonCollection', '-threads', '8'])

        for doc in tqdm(dataset.docs_iter(), total=dataset.docs_count(),
                        disable=self.disable_tqdm, desc='Indexing'):
            document = JsonCollectionDocument.fromFields(doc.doc_id, self.doc_text(doc))
            indexer.addJsonDocument(document)

        indexer.close(True)

    def import_ciff(self, ciff_file: Path, index_name: Path):
        subprocess.run(['CiffImporter', str(ciff_file), str(index_name)])

    def export_ciff(self, index_name: Path, ciff_file: Path):
        subprocess.run(['ExportAnseriniLuceneIndex', '-output', str(ciff_file), '-index', str(index_name),
                        '-description', 'Anserini index'])

    @staticmethod
    def tokenize(text: str):
        glue = String(" ")
        return String.join(glue, AnalyzerUtils.analyze(text))


class GeeseDBIndex(Index):

    delete_chars = {'', '.', ',', "'", '"', '’', '‘', '“', '”', '-', '—', '(', ')', '[', ']', '{', '}',
                    '<', '>', ':', ';', '?', '!', '/', "\\", '=', '&', '$', '€'}

    def index_collection(self, index_name: Path, dataset: Dataset):
        nltk.download('punkt')

        index_name.mkdir(parents=True, exist_ok=True)

        all_tokens = Counter()
        tokenized_docs = []

        with Pool(processes=16) as pool:
            for doc_id, term_counts in tqdm(pool.imap_unordered(self.tokenize_doc, dataset.docs_iter()),
                                            total=dataset.docs_count(), disable=self.disable_tqdm,
                                            desc='Indexing (pass 1)'):
                tokenized_docs.append((doc_id, dict(term_counts)))
                all_tokens.update(term_counts.keys())

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir)

            dict_df = pd.DataFrame(all_tokens.items(), columns=['string', 'df']).sort_values('string', ignore_index=True)
            dict_df['term_id'] = dict_df.index
            dict_df = dict_df[['term_id', 'string', 'df']]

            dict_df.to_csv(tmp_path / 'dict.csv', header=False, index=False, sep='|')

            token_lookup = dict(zip(dict_df['string'], dict_df['term_id']))

            with open(tmp_path / 'docs.csv', 'w', newline='') as docs_file, \
                 open(tmp_path / 'term_doc.csv', 'w', newline='') as term_doc_file:
                doc_writer = csv.writer(docs_file, delimiter='|')
                term_doc_writer = csv.writer(term_doc_file, delimiter='|')

                # doc_writer.writerow(['collection_id', 'doc_id', 'len'])
                # term_doc_writer.writerow(['term_id', 'doc_id', 'tf'])

                tokenized_docs = tqdm(tokenized_docs, disable=self.disable_tqdm, desc='Indexing (pass 2)')
                for doc_id, (collection_id, tokens) in enumerate(tokenized_docs):
                    doclength = sum(tokens.values())

                    doc_writer.writerow([collection_id, doc_id, doclength])

                    for token, tf in tokens.items():
                        term_doc_writer.writerow([token_lookup[token], doc_id, tf])

            indexer = FullTextFromCSV(
                database=f'{index_name}/index.db',
                # docs_file=f"'{tmp_path / 'docs.csv'}'",
                # term_dict_file=f"'{tmp_path / 'dict.csv'}'",
                # term_doc_file=f"'{tmp_path / 'term_doc.csv'}'",
                docs_file=tmp_path / 'docs.csv',
                term_dict_file=tmp_path / 'dict.csv',
                term_doc_file=tmp_path / 'term_doc.csv',
            )
            indexer.load_data()

        close_connection()

    def import_ciff(self, ciff_file: Path, index_name: Path):
        index_name.mkdir(parents=True, exist_ok=True)

        with tempfile.TemporaryDirectory() as tmp_dir, \
                open((tmp_path := Path(tmp_dir)) / 'docs.csv', 'w', newline='') as docs_file, \
                open(tmp_path / 'term_doc.csv', 'w', newline='') as term_doc_file, \
                open(tmp_path / 'dict.csv', 'w', newline='') as dict_file, \
                CiffReader(ciff_file) as reader:
            doc_writer = csv.writer(docs_file, delimiter='|')
            term_doc_writer = csv.writer(term_doc_file, delimiter='|')
            dict_writer = csv.writer(dict_file, delimiter='|')

            postings_lists = tqdm(reader.read_postings_lists(), total=reader.header.num_postings_lists,
                                  disable=self.disable_tqdm, desc='Posting lists')
            for term_id, postings_list in enumerate(postings_lists):
                dict_writer.writerow([term_id, postings_list.term, postings_list.df])

                docid = 0
                for posting in postings_list.postings:
                    docid += posting.docid
                    term_doc_writer.writerow([term_id, docid, posting.tf])

            doc_records = tqdm(reader.read_documents(), total=reader.header.num_docs,
                               disable=self.disable_tqdm, desc='Documents')
            for doc_record in doc_records:
                doc_writer.writerow([doc_record.collection_docid, doc_record.docid, doc_record.doclength])

            indexer = FullTextFromCSV(
                database=f'{index_name}/index.db',
                docs_file=tmp_path / 'docs.csv',
                term_dict_file=tmp_path / 'dict.csv',
                term_doc_file=tmp_path / 'term_doc.csv',
            )
            indexer.load_data()

        close_connection()

    def export_ciff(self, index_name: Path, ciff_file: Path):
        ToCiff(database=f'{index_name}/index.db', ciff=str(ciff_file), verbose=self.verbose)
        close_connection()

    @staticmethod
    def tokenize(text):
        return [token.lower() for token in word_tokenize(text)]

    def tokenize_doc(self, doc):
        term_counts = Counter(token for token in self.tokenize(self.doc_text(doc))
                              if token not in self.delete_chars)

        return doc.doc_id, dict(term_counts)
