import subprocess
import tempfile
from pathlib import Path
from warnings import warn

import pandas as pd
from tqdm import tqdm

import pyterrier as pt
from pyterrier.io import read_results
from pyterrier.batchretrieve import BatchRetrieveBase
from pyterrier.model import coerce_queries_dataframe

from geesedb.search import Searcher


def PyTerrierBatchRetrieve(index_location: Path, num_results: int = 1000, b: float = 0.4, k1: float = 0.9, **kwargs):
    return pt.BatchRetrieve(
        pt.IndexFactory.of(f'{index_location}/data.properties'),
        num_results=num_results,
        wmodel='BM25', properties={'termpipelines': ''},
        controls={'bm25.b': b, 'bm25.k_1': k1, 'bm25.k_3': 1.0},
        **kwargs,
    )


class GeeseDBBatchRetrieve(BatchRetrieveBase):
    """Allows retrieval from a GeeseDB index."""

    def __init__(self, index_location: Path, num_results: int = 1000, b: float = 0.4, k1: float = 0.9, **kwargs):
        super().__init__(**kwargs)
        self.index_location = index_location
        self.num_results = num_results

        self.searcher = Searcher(database=str(self.index_location / 'index.db'), return_type='list',
                                 retrieval_method='BM25_robertson', b=b, k1=k1, n=num_results)

    def transform(self, queries):
        results = []
        if not isinstance(queries, pd.DataFrame):
            warn('.transform() should be passed a dataframe. Use .search() to execute a single query.',
                 DeprecationWarning, 2)
            queries = coerce_queries_dataframe(queries)

        if 'docno' in queries.columns:
            raise ValueError(f'{self} does not support reranking')

        for row in tqdm(queries.itertuples(), desc=self.name, total=queries.shape[0], unit='q',
                        disable=not self.verbose):
            qid = str(row.qid)
            query = row.query

            # Required because some of the tokenizers do not remove apostrophes, which will result in SQL errors
            escaped_query = query.replace("'", "''")

            hits = self.searcher.search_topic(escaped_query)

            for rank, (docno, score) in enumerate(hits):
                results.append([qid, query, docno, rank, score])

        return pd.DataFrame(results, columns=['qid', 'query', 'docno', 'rank', 'score'])


class AnseriniBatchRetrieve(BatchRetrieveBase):
    """Allows retrieval from an Anserini index."""

    def __init__(self, index_location: Path, num_results: int = 1000, b: float = 0.4, k1: float = 0.9, **kwargs):
        super().__init__(**kwargs)
        self.index_location = index_location
        self.num_results = num_results
        self.b = b
        self.k1 = k1

    def transform(self, queries):
        if not isinstance(queries, pd.DataFrame):
            warn('.transform() should be passed a dataframe. Use .search() to execute a single query.',
                 DeprecationWarning, 2)
            queries = coerce_queries_dataframe(queries)

        if 'docno' in queries.columns:
            raise ValueError(f'{self} does not support reranking')

        with tempfile.TemporaryDirectory() as tmpdir:
            query_file = Path(tmpdir) / 'queries.tsv'
            run_file = Path(tmpdir) / 'anserini.run'

            queries[['qid', 'query']].to_csv(query_file, header=False, index=False, sep='\t')

            subprocess.run(['SearchCollection', '-index', str(self.index_location), '-hits', str(self.num_results),
                            '-bm25', '-bm25.b', str(self.b), '-bm25.k1', str(self.k1),
                            '-keepstopwords', '-pretokenized', '-stemmer', 'none',
                            '-topicreader', 'TsvString', '-topics', str(query_file),
                            '-output', str(run_file)])

            return read_results(str(run_file))
